#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : MeUtils.
# @File         : __init__.py
# @Time         : 2021/1/31 10:20 下午
# @Author       : yuanjie
# @Email        : meutils@qq.com
# @Software     : PyCharm
# @Description  : python meutils/clis/__init__.py


import typer

from meutils.pipe import *

cli = typer.Typer(name="llm4gpt CLI")


@cli.command(help="help")  # help会覆盖docstring
def clitest(name: str='TEST'):
    """

    @param name: name
    @return:
    """
    typer.echo(f"Hello {name}")


if __name__ == '__main__':
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : utils
# @Time         : 2023/4/20 12:50
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import torch
from transformers import AutoTokenizer, AutoModel, AutoConfig

from meutils.pipe import *


def auto_configure_device_map(num_gpus: int) -> Dict[str, int]:
    # transformer.word_embeddings 占用1层
    # transformer.final_layernorm 和 lm_head 占用1层
    # transformer.layers 占用 28 层
    # 总共30层分配到num_gpus张卡上
    num_trans_layers = 28
    per_gpu_layers = 30 / num_gpus

    # bugfix: 在linux中调用torch.embedding传入的weight,input不在同一device上,导致RuntimeError
    # windows下 model.device 会被设置成 transformer.word_embeddings.device
    # linux下 model.device 会被设置成 lm_head.device
    # 在调用chat或者stream_chat时,input_ids会被放到model.device上
    # 如果transformer.word_embeddings.device和model.device不同,则会导致RuntimeError
    # 因此这里将transformer.word_embeddings,transformer.final_layernorm,lm_head都放到第一张卡上
    device_map = {'transformer.word_embeddings': 0,
                  'transformer.final_layernorm': 0, 'lm_head': 0}

    used = 2
    gpu_target = 0
    for i in range(num_trans_layers):
        if used >= per_gpu_layers:
            gpu_target += 1
            used = 0
        assert gpu_target < num_gpus
        device_map[f'transformer.layers.{i}'] = gpu_target
        used += 1

    return device_map


def llm_load(model_name_or_path="THUDM/chatglm-6b", device='cpu', device_map: Optional[Dict[str, int]] = None,
             **kwargs):
    tokenizer = AutoTokenizer.from_pretrained(model_name_or_path, trust_remote_code=True)
    model = AutoModel.from_pretrained(model_name_or_path, trust_remote_code=True, **kwargs)

    if torch.cuda.is_available() and device.lower().startswith("cuda"):
        # 根据当前设备GPU数量决定是否进行多卡部署
        num_gpus = torch.cuda.device_count()
        if num_gpus < 2 and device_map is None:
            model = model.half().cuda()
        else:
            from accelerate import dispatch_model
            # 可传入device_map自定义每张卡的部署情况
            if device_map is None:
                device_map = auto_configure_device_map(num_gpus)

            model = dispatch_model(model, device_map=device_map).half().cuda()

    else:
        model = model.float().to(device)

    return model.eval(), tokenizer


def llm_load4chat(model_name_or_path="THUDM/chatglm-6b", device='cpu', stream=True, **kwargs):
    model, tokenizer = llm_load(model_name_or_path, device, **kwargs)
    if stream:
        return partial(model.stream_chat, tokenizer=tokenizer)
    else:
        return partial(model.chat, tokenizer=tokenizer)


if __name__ == '__main__':
    a = llm_load("/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : ann4qa
# @Time         : 2023/4/24 18:10
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from meutils.pipe import *
from llm.applications.qa import QA


class ANN4QA(QA):

    def __init__(self, chat_func):
        super().__init__(chat_func)

    # def qa(self, query, knowledge_base='', **kwargs):
    #     pass


if __name__ == '__main__':
    from llm.utils import llm_load4chat

    chat_func = llm_load4chat(
        model_name_or_path="/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm",
        device='mps',
    )

    qa = ANN4QA(chat_func=chat_func)

    for i, _ in qa(query='1+1'):
        print(i)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : FaissANN
# @Time         : 2023/4/20 17:22
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from langchain.vectorstores import FAISS
from langchain.embeddings.huggingface import HuggingFaceEmbeddings

# ME
from meutils.pipe import *


class FaissANN(object):
    def __init__(self, folder_path: str = None,
                 index_name: str = "index",
                 model_name_or_path="shibing624/text2vec-base-chinese"):
        self.embeddings = HuggingFaceEmbeddings(model_name=model_name_or_path)
        self.folder_path = folder_path
        self.faiss_ann = self.load_local(folder_path, index_name)  # 加载已存在的索引

    def add_texts(self, texts, metadatas: Optional[List[dict]] = None):  # todo: 增加进度条
        self.faiss_ann = FAISS.from_texts(texts, self.embeddings, metadatas)  # metadatas = [{'source': 'xx'}]

    def update(self, target: FAISS, index_name=None):
        self.faiss_ann.merge_from(target)  # Add the target FAISS to the current one.
        if index_name:
            logger.info('保存新的索引文件')
            self.faiss_ann.save_local(self.folder_path, index_name)

    def load_local(self, folder_path=None, index_name: str = "index"):
        if folder_path and Path(folder_path).is_dir():
            return FAISS.load_local(folder_path, self.embeddings, index_name)



from langchain.text_splitter import CharacterTextSplitter
import re
from typing import List


class ChineseTextSplitter(CharacterTextSplitter):
    def __init__(self, pdf: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.pdf = pdf

    def split_text(self, text: str) -> List[str]:
        if self.pdf:
            text = re.sub(r"\n{3,}", "\n", text)
            text = re.sub('\s', ' ', text)
            text = text.replace("\n\n", "")
        sent_sep_pattern = re.compile('([﹒﹔﹖﹗．。！？]["’”」』]{0,2}|(?=["‘“「『]{1,2}|$))')  # del ：；
        sent_list = []
        for ele in sent_sep_pattern.split(text):
            if sent_sep_pattern.match(ele) and sent_list:
                sent_list[-1] += ele
            elif ele:
                sent_list.append(ele)
        return sent_list





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : Question2Answer
# @Time         : 2023/4/21 12:25
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import types
from meutils.pipe import *
from meutils.decorators import clear_cuda_cache


class QA(object):

    def __init__(self, chat_func, prompt_template=None):
        self.chat_func = chat_func

        self.history = []
        self.knowledge_base = None

        self.prompt_template = prompt_template
        if prompt_template is None:
            self.prompt_template = self.default_document_prompt

    def __call__(self, **kwargs):
        return self.qa(**kwargs)

    @abstractmethod  # 重写
    def qa(self, query, knowledge_base='', **kwargs):
        return self._qa(query, knowledge_base, **kwargs)

    @clear_cuda_cache
    def _qa(self, query, knowledge_base='', max_turns=1):
        if knowledge_base:
            self.knowledge_base = knowledge_base
            query = self.prompt_template.format(context=knowledge_base, question=query)

        result = self.chat_func(query=query, history=self.history[-max_turns:])

        if isinstance(result, types.GeneratorType):
            return self._stream(result)
        else:  # list(self._stream(result)) 想办法合并
            response, history = result
            # self.history_ = history  # 历史所有
            self.history += [[None, response]]  # 置空知识

        return response

    def _stream(self, result):  # yield > return
        response = None
        for response, history in tqdm(result, desc='Stream', ascii=True, ncols=100):
            yield response, history
        # self.history_ = history  # 历史所有
        self.history += [[None, response]]  # 置空知识

    @property
    def default_document_prompt(self):
        prompt_template = """
            基于以下已知信息，简洁和专业的来回答用户的问题。
            如果无法从中得到答案，请说 "根据已知信息无法回答该问题" 或 "没有提供足够的相关信息"，不允许在答案中添加编造成分，答案请使用中文。
            已知内容: {context}
            问题: {question}
            """.strip()

        return prompt_template

    def set_chat_kwargs(self, **kwargs):
        self.chat_func = partial(self.chat_func, **kwargs)


if __name__ == '__main__':
    from llm.utils import llm_load4chat

    chat_func = llm_load4chat(
        model_name_or_path="/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm",
        device='mps')

    qa = QA(chat_func=chat_func)

    for i, _ in qa(query='提取关键词', knowledge_base='我今天去听了周杰伦的演唱会'):
        print(i, flush=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : sota_embedding
# @Time         : 2023/4/21 11:56
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :

from sentence_transformers import SentenceTransformer
from meutils.docarray_ import Document, DocumentArray


# 增加 docarray v2
# 增加 ann后端
class DocEmbedding(object):
    def __init__(self, model_name_or_path="shibing624/text2vec-base-chinese", device=None):
        """
            disk_cache()(DocEmbedding().encode)
        :param model_name_or_path:
        :param device:
        """
        self.st = SentenceTransformer(model_name_or_path, device)

    def __call__(self, *args, **kwargs):
        return self.encode(*args, **kwargs)

    def encode(self, sentences, batch_size=32, show_progress_bar=False, normalize_embeddings=False,
               return_document=False):

        if isinstance(sentences, str):
            sentences = [sentences]

        embeddings = self.st.encode(
            sentences,
            batch_size=batch_size,
            show_progress_bar=show_progress_bar,
            normalize_embeddings=normalize_embeddings
        )
        if return_document:
            da = DocumentArray.empty(len(sentences))
            da.texts = sentences
            da.embeddings = embeddings
            return da
        return embeddings



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : Question2Answer
# @Time         : 2023/4/21 12:25
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import types
from meutils.pipe import *
from meutils.docarray_ import DocumentArray
from meutils.decorators import clear_cuda_cache
from meutils.request_utils.crawler import Crawler


class Question2Answer(object):

    def __init__(self, chat_func, prompt_template=None):
        self.chat_func = chat_func
        # self.query_embedd = lru_cache()(query_embedd)  # 缓存
        # self.docs = docs

        self.history = []

        self.prompt_template = prompt_template
        if prompt_template is None:
            self.prompt_template = self.default_document_prompt

    @abstractmethod
    def qa(self):
        raise NotImplementedError("overwrite method!!!")

    def search4qa(self):
        pass

    def crawler4qa(self, query,
                   url="https://top.baidu.com/board?tab=realtime",
                   xpath='//*[@id="sanRoot"]/main/div[2]/div/div[2]/div[*]/div[2]/a/div[1]//text()', **kwargs):
        knowledge_base = Crawler(url).xpath(xpath)

        return self._qa(query, knowledge_base, **kwargs)

    def ann4qa(self, query, query_embedd=None, da: DocumentArray = None, topk=3, **kwargs):

        # ann召回知识
        v = query_embedd(query)
        knowledge_base = da.find(v, topk=topk)[0].texts  # [:, ('text', 'scores__cosine__value')]

        return self._qa(query, knowledge_base, **kwargs)

    @clear_cuda_cache
    def _qa(self, query, knowledge_base='', max_turns=1, print_knowledge_base=False):
        if knowledge_base:
            query = self.prompt_template.format(context=knowledge_base, question=query)

            if print_knowledge_base:
                pprint({'knowledge_base': knowledge_base})

        result = self.chat_func(query=query, history=self.history[-max_turns:])

        if isinstance(result, types.GeneratorType):
            return self._stream(result)
        else:  # list(self._stream(result)) 想办法合并
            response, history = result
            # self.history_ = history  # 历史所有
            self.history += [[None, response]]  # 置空知识

        return response

    def _stream(self, result):  # yield > return
        response = None
        for response, history in tqdm(result, desc='Stream'):
            yield response, history
        # self.history_ = history  # 历史所有
        self.history += [[None, response]]  # 置空知识

    @property
    def default_document_prompt(self):
        prompt_template = """
            基于以下已知信息，简洁和专业的来回答用户的问题。
            如果无法从中得到答案，请说 "根据已知信息无法回答该问题" 或 "没有提供足够的相关信息"，不允许在答案中添加编造成分，答案请使用中文。
            已知内容:
            {context}
            问题:
            {question}
            """.strip()

        return prompt_template


if __name__ == '__main__':
    from llm.utils import llm_load

    model, tokenizer = llm_load("/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm")
    qa = Question2Answer(
        # chat_func=partial(model.stream_chat, tokenizer=tokenizer),
        chat_func=partial(model.chat, tokenizer=tokenizer),

    )

    # for i, _ in qa._qa('1+1'):
    #     print(i, flush=True)
    print(qa._qa('1+1'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : AI.  @by PyCharm
# @File         : chatllm
# @Time         : 2023/4/20 15:46
# @Author       : betterme
# @WeChat       : meutils
# @Software     : PyCharm
# @Description  :
import types
from langchain.llms.base import LLM
from langchain.llms.utils import enforce_stop_tokens

# ME
from meutils.pipe import *


class ChatLLM(LLM):
    """
    from llm.utils import llm_load

    model, tokenizer = llm_load("/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm")
    glm = ChatLLM()
    glm.chat_func = model.chat # partial(self.chat_func, **kwargs)
    """
    chat_func: Callable = None
    history = []
    max_turns = 3

    @property
    def _llm_type(self) -> str:
        return "ChatLLM"

    # todo: https://github.com/hwchase17/langchain/issues/2415 增加流失

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:

        result = self.chat_func(query=prompt, history=self.history[-self.max_turns:])
        response = history = None
        if isinstance(result, types.GeneratorType):
            for response, history in tqdm(result, desc='Stream'):
                yield response, history
        else:
            response, history = result

        if stop:
            response = enforce_stop_tokens(response, stop)
        self.history = history  # 历史所有         self.history += [[None, response]]

        return response

    def set_chat_kwargs(self, **kwargs):
        self.chat_func = partial(self.chat_func, **kwargs)


if __name__ == '__main__':
    from llm.utils import llm_load

    model, tokenizer = llm_load("/Users/betterme/PycharmProjects/AI/CHAT_MODEL/chatglm")
    glm = ChatLLM()
    glm.chat_func = partial(model.chat, tokenizer=tokenizer)
    glm.chat_func = partial(model.stream_chat, tokenizer=tokenizer)
